from __future__ import annotations
from dataclasses import dataclass
import os
from typing import Any, ByteString, Dict, List, Callable
import typing
import re
import itertools
import json
from itertools import cycle
from pathlib import PurePath
import plotly.graph_objects as go
from plotly.graph_objects import Figure
import plotly.express as px
from plotly.subplots import make_subplots
import pandas as pd
import plotly.io as pio
from enum import Enum
import functools

class COLORS:
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    WARNING = '\033[93m'
    PURPLE = '\033'
    ENDC = '\033[0m'
    LIGHT_BLUE = '\033[34m'

RegexResult = List[Dict[str, str]]
class ExplainAnalyzeData:
    raw_data: RegexResult
    compile_time: float
    codegen_time: float
    optimization_time: float
    val_int_call_time: float
    instruction_count: int
    total_time: float


    def __init__(self):
        self.raw_data = []
        self.compile_time = 0
        self.codegen_time = 0
        self.val_int_call_time = 0
        self.total_time = 0
        self.optimization_time = 0
        self.instruction_count = 0
        pass

    def add_group(self, regex_groups: typing.Dict[str, str]):
        self.raw_data.append(regex_groups)
        self.codegen_time += float(regex_groups['codegen'] or 0)
        self.compile_time += float(regex_groups['compile'] or 0)
        self.val_int_call_time += float(regex_groups['val_int_calls'] or 0)
        self.optimization_time += float(regex_groups['opt'] or 0)
        self.instruction_count += int(regex_groups['instruction_count'] or 0)

    def set_total_time(self, time: float):
        self.total_time = time

    def add(self, data: ExplainAnalyzeData):
        self.compile_time += data.compile_time
        self.codegen_time += data.codegen_time
        self.val_int_call_time += data.val_int_call_time
        self.total_time += data.total_time
        self.optimization_time += data.optimization_time
        self.instruction_count += data.instruction_count
    
    def divide(self, n: int):
        self.compile_time /= n
        self.codegen_time /= n
        self.val_int_call_time /= n
        self.total_time /= n
        self.optimization_time /= n
        self.instruction_count /= n

    def num_rows(self) -> int:
        return sum([int(raw_data['rows'])*int(raw_data['loops']) for raw_data in self.raw_data if raw_data['name'].startswith('Table scan')])

    def from_json(
        file: str, 
        key_codegen: str="codegen", 
        key_compile:str="compile",
        key_val_int:str="val_int_call_time",
        key_total_time:str="total_time",
        key_opt:str="opt",
        key_instruction_count:str="instruction_count"
    ) -> ExplainAnalyzeData:
        data = ExplainAnalyzeData()
        with open(file, 'r') as f:
            json_data = json.load(f)
            data.codegen_time += json_data[key_codegen] or 0
            data.compile_time += json_data[key_compile] or 0
            data.val_int_call_time += json_data[key_val_int] or 0
            data.total_time += json_data[key_total_time] or 0
            data.optimization_time += (json_data[key_opt] or 0) if key_opt in json_data else 0
            data.instruction_count += (json_data[key_instruction_count] or 0) if key_instruction_count in json_data else 0
        return data

    def __str__(self) -> str:
        return f"ExplainAnalyzeData(compile_time={self.compile_time}, codegen_time={self.codegen_time}, val_int_call_time={self.val_int_call_time}, total_time={self.total_time}, opt_time={self.optimization_time}, instr_count={self.instruction_count})"

    def is_empty(self) -> bool:
        return len(self.raw_data) == 0


@dataclass
class WhereAnalyzeData:
    scale: str
    type: str
    size: int
    data: ExplainAnalyzeData

@dataclass
class WhereBenchmarkData:
    scale: str
    type: str
    size: int
    data: float

class AnalyzeDataIndex(object):
    X = 0
    TOTAL_TIME = 1
    COMPILE_TIME = 2
    CODEGEN_TIME = 3
    VAL_INT_TIME = 4
    OPT_TIME = 5
    INSTR_COUNT = 6
    EXPLAIN_ANALYZE_DATA = 7


def read_file(path: str, filename: str, split: bool = True) -> List[str]:
    path = str(PurePath(path, filename))
    if not os.path.exists(path):
        return None
    with open(path, 'r') as f:
        if split:
            return list(itertools.chain(*[line.split("\\n") for line in f.read().split("\n")]))
        else:
            return f.readlines()

def split_newline(lines: ByteString | str):
    if not isinstance(lines, str):
        lines = str(lines, 'utf-8')

    return list(itertools.chain(*[line.split("\\n") for line in lines.split("\n")]))

def extract_time(lines: List[str]) -> ExplainAnalyzeData:
    data = ExplainAnalyzeData()
    prog = re.compile('(?P<indent>[ ]*)[->]+ (?P<name>[a-zA-Z-]+:?(?: [a-zA-Z-<>_]+){0,}){1}((\[[a-zA-Z0-9]+\]\((codegen=(?P<codegen>\d+\.\d+) compile=(?P<compile>\d+\.\d+)( opt=(?P<opt>\d+.\d+))?))|).*(time_spent_on_val_int_calls=(?P<val_int_calls>\d+\.\d+)( instruction_count=(?P<instruction_count>\d+))?).*\(actual time=(?P<actual_time_from>\d+\.\d+)..(?P<actual_time_to>\d+\.\d+) rows=(?P<rows>\d+) loops=(?P<loops>\d+)')
    for line in lines:
        match = prog.match(line)
        if match is None:
            continue
        groups = match.groupdict()

        data.add_group(groups)
        
        # Extract the time from the last line
        if data.total_time == 0 and (lines[-1].isnumeric() or lines[-2].isnumeric()):
            data.set_total_time((float(lines[-2]) if len(lines[-1]) == 0 else float(lines[-1]))*1000)

    return data


def extract_benchmark_data(path: str = 'data/mysql/normal/'):
    x = []
    y = []
    for i in range(0, 22 + 1):
        lines = read_file(path, f"{i}.benchmark.sql")
        if lines is None:
            continue
        
        x.append(i)
        y.append(float(lines[0]))
    return (x, y)

ExtractAnalyzeData = typing.Tuple[List[int], List[float], List[float], List[float], List[float], List[float], List[int], List[ExplainAnalyzeData]]
def extract_analyze_data(path: str = 'data/mysql/normal/') -> ExtractAnalyzeData:
    x = []
    total_time = []
    compile_time = []
    codegen_time = []
    val_int_time = []
    opt_time = []
    instruction_count = []
    analyze_data: List[ExplainAnalyzeData] = []
    for i in range(0, 22 + 1):
        lines = read_file(path, f"{i}.analyze.metadata.sql", split=False)
        if lines is None:
            continue
        data = json.loads(lines[0])
        
        x.append(i)
        total_time.append(data["total_time"])
        compile_time.append(data["compile"])
        codegen_time.append(data["codegen"])
        val_int_time.append(data["val_int_call_time"])
        opt_time.append(data["opt"] if "opt" in data else 0)
        instruction_count.append(data["instruction_count"] if "instruction_count" in data else 0)

        with open(str(PurePath(path, f'{i}.analyze.sql')), 'r') as f:
            raw_explain_analyze_data = extract_time(split_newline(f.readlines()[1]))
            analyze_data.append(raw_explain_analyze_data)

    return (x, total_time, compile_time, codegen_time, val_int_time, opt_time, instruction_count, analyze_data)

def extract_where_analyze_data(scale: str, path: str = 'data/mysql/m_tpch3_v1_jit/') -> Dict[str, WhereAnalyzeData]:
    files = [PurePath(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f)) and 'analyzedata' in f]

    data = {}
    for f in files:
        comps = str(f).split('.')
        if len(comps) != 5:
            print(f"the file '{f}' is not a valid where-file")
            return None
        
        type = comps[-3]
        size = comps[-2]
        d = WhereAnalyzeData(scale, type, int(size), ExplainAnalyzeData.from_json(f))

        with open(str(f).replace('analyzedata', 'analyze'), 'r') as f:
            # print(split_newline(f.readlines()[1]))
            raw_explain_analyze_data = extract_time(split_newline(f.readlines()[1]))
            d.data.raw_data = raw_explain_analyze_data.raw_data
        
        data[f"{type}.{size}"] = d
    return data

def extract_where_benchmark_data(scale: str, path: str = 'data/mysql/m_tpch3_v1_jit/') -> Dict[str, WhereAnalyzeData]:
    files = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f)) and 'benchmark' in f]

    data = {}
    for f in files:
        comps = str(f).split('.')
        if len(comps) != 5:
            print(f"the file '{f}' is not a valid where-file")
            return None
        type = comps[2]
        size = comps[3]

        lines = read_file(path, f)
        if lines is None:
            continue
        
        d = WhereBenchmarkData(scale, type, int(size), float(lines[0]))
        data[f"{type}.{size}"] = d
    return data

Data = typing.Tuple[List[int], List[float]]
def extract_jit_comparison(njit: Data, jit: Data):
    x = []
    y1 = []

    z_njit = zip(njit[0], njit[1])
    z_jit = zip(jit[0], jit[1])

    def cmp(a: int, b: int):
        val = ((a-b)/b*100 if a > b else (b-a)/b*100)
        # print(f"A: {a}, B: {b}, VAL: {val if a > b else -val}")
        return val if a > b else -val

    for i, (x_njit, y_njit) in enumerate(z_njit):
        x.append(x_njit)

        y1.append(0)
        for j, (x_jit, y_jit) in enumerate(z_jit):
            if x_jit == x_njit:
                # print(i, x_njit, y_njit, y_jit, cmp(y_njit, y_jit))
                y1[i] = cmp(y_njit, y_jit)
                break 
    
    return (x, y1)


def extract_data_times_by_key(path: str = 'data/mysql/m_tpch1', func: Callable[[ExplainAnalyzeData], float] = lambda x: x.compile_time):
    x = []
    y = []
    for i in range(0, 22 + 1):
        lines = read_file(path, f"{i}.analyze.sql")
        
        x.append(i)
        y.append(0)
        if lines is None:
            continue
        data = extract_time(lines)
        if data.is_empty():
            continue
        
        y[i] = func(data)
        
    return (x, y)

def list_subtraction(lists: List[typing.Tuple[List, List]]):
    if len(lists) == 0:
        return []
    elif len(lists) == 1:
        return lists[0][1]

    data = {}
    for list_item in lists:
        xs = list_item[0]
        ys = list_item[1]
        for i, x in enumerate(xs):
            if xs[i] not in data:
                data[xs[i]] = ys[i]
            else:
                data[xs[i]] -= ys[i]

    x = [value[0] for value in sorted(list(data.items()), key=lambda x: x[0])]
    y = [value[1] for value in sorted(list(data.items()), key=lambda x: x[0])]
    return (x, y)


def extract_num_joins(path: str):
    x = []
    y = []
    for i in range(0, 22 + 1):
        lines = read_file(path, f"{i}.analyze.sql")
        if lines is None:
            continue
        
        line = '\n'.join(lines)
        prog = re.compile('(?i)(join)')
        match = prog.findall(line)
        if match is None:
            print("No matches")
            return None
        
        x.append(i)
        y.append(len(match))
    return (x, y)

def table_figure_to_csv(fig: pd.Figure, name: str):
    pd.DataFrame(
        data=fig.data[0]['cells']['values'],
        index=fig.data[0]['header']['values']
    ).transpose().to_csv(f'img/{name.removesuffix(".csv")}.csv', index=False, header=True)

def figure_to_image(fig: pd.Figure, name: str):
    pio.write_image(fig, f'img/{name.removesuffix(".png")}.png', format='png', scale=5)


def geometric_mean(l: List[float]):
    if len(l) == 0:
        return 0
    product: float = functools.reduce(lambda a, b: (a * 1000) * (b * 1000), l)/1000000
    return product**(1/len(l))

def calc_speedup(a: float | int | List[float], b: float | int | List[float]):
    def cmp(a: float | int, b: float | int):
        if a == 0 or b == 0:
            return -1

        val = ((a-b)/b*100 if a > b else (b-a)/b*100)
        return val if a > b else -val

    if isinstance(a, list) and isinstance(b, list):
        return [cmp(val, b[i]) for i, val in enumerate(a) if i < len(b)]
    else:
        return cmp(a, b)

def render(
    data_list: typing.Tuple[typing.Tuple[typing.List, typing.List], str], 
    title: str = 'TPC-H Queries, Scale 10',
    xaxis_title= 'TPC-H Query',
    yaxis_title= 'Avg execution time (ms)',
    log: bool = False,
    log_x: bool = False,
    exclude_x: typing.List = [],
    x_as_str: bool = False,
    grayscale: bool = True,
    line: bool = False,
    barmode: str = None,
    dtick: int = 1,
    yrange: List = None,
    xrange: List = None,
    show: bool = False,
    width: int = None,
    legend_orientation: str = None,
    bar_options: dict = {},
    **kwargs,
):
    pattern_shapes = cycle(['.', 'x', '+', '|', '/'])
    symbol_shapes = cycle(['circle', 'cross', 'diamond', 'triangle-up', 'star'])
    dashes = cycle(['solid', 'dot', 'dash', 'dashdot', 'longdash'])
    colors = cycle(list(set(px.colors.sequential.gray)))
    next(colors) # The first one is a bad color

    # Exclude values of x
    # NB! To use this feature, make sure each entry have its own copy of x and y list
    for value in exclude_x:
        # index = -1
        for (data, name) in data_list:
            if value in data[0]:
                index = data[0].index(value)
                data[0].pop(index)
                data[1].pop(index)


    if x_as_str:
        data_list = [((list(map(lambda x: str(x), data[0])), data[1]), name) for (data, name) in data_list]

    objects = []
    for data_item in data_list:
        data = data_item[0]
        name = data_item[1]
        options: dict = {} if len(data_item) <= 2 else data_item[2]
        options = {**bar_options, **options, "name": name, "y": data[1], "x": data[0], "marker_pattern_shape": next(pattern_shapes)}
        obj = go.Bar(**options)
        if line:
            obj = go.Scatter(y=data[1], x=data[0], name=name, mode='lines+markers', marker=dict(symbol=next(symbol_shapes)), line=dict(dash=next(dashes)))

        if grayscale:
            obj.update(marker_color=next(colors))
        objects.append(obj)

    fig = go.Figure(
        data=objects,
        layout_title_text=title,
    )

    if "legend" not in kwargs:
        kwargs["legend"] = dict() 
    kwargs["legend"]["orientation"] = legend_orientation

    fig.update_layout(
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
        width=width,
        **kwargs
    )

    if barmode:
        fig.update_layout(barmode=barmode)

    if yrange:
        fig.update_layout(yaxis_range=yrange)

    if xrange:
        fig.update_layout(xaxis_range=xrange)

    if dtick is not None:
        fig.update_xaxes(dtick=dtick)

    if log:
        fig.update_yaxes(type="log")
    
    if log_x:
        fig.update_xaxes(type="log")

    if show:
        fig.show()

    return fig


def render_table(headers: List[str], values: List[List[any]], format_float: str | None = None):
    columns = [[format_float.format(value) if format_float is not None and isinstance(value, float) else  value for value in column] for column in values]


    fig = go.Figure(data=[go.Table(
        header=dict(values=headers,
                    # line_color='darkslategray',
                    # fill_color='lightskyblue',
                    align='left'),
        cells=dict(values=columns,
                #    line_color='darkslategray',
                #    fill_color='lightcyan',
                align='left'))
    ])

    return fig

def render_subplots(figs: List[Figure], cols: int, rows: int, **kwargs):
    fig = make_subplots(rows=rows, cols=cols, **kwargs)

    i = 0
    for f in figs:
        col = int((i%cols) + 1)
        row = int((i//cols) + 1)
        fig.add_trace(f.data[0], row=row, col=col)
        i += 1

    return fig

# -------- POSTGRES SPECIFIC STUFF -----------

# extract_jit_time extracts the jit time from a Postgres Explain Analyze output 
def extract_jit_time(lines: List[str]):
    times = {}
    line = '\n'.join(lines)
    prog = re.compile('[.\n a-zA-Z0-9:\(\)=<>\[\]\'},{_"$%#!*~-]*JIT:\n   Functions: (\d+)[\w: \n]+(?P<functions>true|false).*(?P<inlining>true|false).*(?P<optimization>true|false).*(?P<expressions>true|false)\n[a-zA-Z: ]+(?P<deforming>[0-9\.]+)[a-zA-Z, ]+(?P<generation_t>[0-9\.]+)[a-zA-Z, ]+(?P<inlining_t>[0-9\.]+)[a-zA-Z, ]+(?P<emission_t>[0-9\.]+)[a-zA-Z, ]+(?P<total_t>[0-9\.]+) ms\n[a-zA-Z:, ]+(?P<execution_t>[0-9\.]+)')
    match = prog.match(line)
    if match is None:
        return None

    d = match.groupdict()
    for key, value in d.items():
        if 'true' in value or 'false' in value:
            continue 
        d[key] = float(value) 
    return d
        
# extract_jit_time extracts the jit time from a Postgres Explain Analyze output 
def extract_jit_times(path: str):
    x = []
    y1 = []
    y2 = []
    y3 = []
    y4 = []
    for i in range(0, 22 + 1):
        lines = read_file(path, f"{i}.analyze.sql")
        if lines is None:
            continue
        times = extract_jit_time(lines)
        if times is None:
            continue
        x.append(i)
        y1.append(times['total_t'])
    
    return (x, y1)

if __name__ == '__main__':
    with open('data/mysql/m_tpch01_where_jit/where.analyze.str.1000.sql', 'r') as f:
        line = f.readlines()[1]
        new_lines = split_newline(line)
        result = extract_time(new_lines)
        print(result)
        